@@ -35,7 +35,6 @@ class _LWRFWebsocketMessage:
     def json(self):
         return json.dumps(self._message)
 
-
 class _LWRFWebsocketMessageItem:
     _item_id = 0
 
@@ -50,44 +49,41 @@ class _LWRFWebsocketMessageItem:
 class LWRFFeatureSet:
 
     def __init__(self):
+        self.link = None
         self.featureset_id = None
         self.name = None
         self.product_code = None
         self.features = {}
 
-    def is_switch(self):
-        return ('switch' in self.features.keys()) and not ('dimLevel' in self.features.keys())
+    def has_feature(self, feature): return feature in self.features.keys()
 
-    def is_light(self):
-        return 'dimLevel' in self.features.keys()
+    def is_switch(self): return (self.has_feature('switch')) and not (self.has_feature('dimLevel'))
+    def is_light(self): return self.has_feature('dimLevel')
+    def is_climate(self): return self.has_feature('targetTemperature')
+    def is_trv(self): return self.has_feature('valveSetup')
+    def is_cover(self): return self.has_feature('threeWayRelay')
+    def is_energy(self): return (self.has_feature('energy')) and (self.has_feature('rssi'))
+    def is_windowsensor(self): return self.has_feature('windowPosition')
+    def is_hub(self): return self.has_feature('buttonPress')
 
-    def is_climate(self):
-        return 'targetTemperature' in self.features.keys()
+    def is_gen2(self): return self.has_feature('upgrade')
+    def reports_power(self): return self.has_feature('power')
+    def has_led(self): return self.has_feature('rgbColor')
 
-    def is_trv(self):
-        return 'valveSetup' in self.features.keys()
+class LWRFFeature:
 
-    def is_cover(self):
-        return 'threeWayRelay' in self.features.keys()
+    def __init__(self):
+        self.featureset = None
+        self.id = None
+        self.name = None
+        self._state = None
 
-    def is_energy(self):
-        return ('energy' in self.features.keys()) and ('rssi' in self.features.keys())
-
-    def is_windowsensor(self):
-        return 'windowPosition' in self.features.keys()
-
-    def is_hub(self):
-        return 'buttonPress' in self.features.keys()
-
-    def is_gen2(self):
-        return 'identify' in self.features.keys()
-
-    def reports_power(self):
-        return 'power' in self.features.keys()
-
-    def has_led(self):
-        return 'rgbColor' in self.features.keys()
+    @property
+    def state(self):
+        return self._state
 
+    async def set_state(self, value):
+        await self.featureset.link.async_write_feature(self.id, value)
 
 class LWLink2:
 
@@ -129,6 +125,7 @@ class LWLink2:
             _LOGGER.debug("async_sendmessage: [contents hidden for security]")
         else:
             _LOGGER.debug("async_sendmessage: Sending: %s", message.json())
+
         await self._websocket.send_str(message.json())
         _LOGGER.debug("async_sendmessage: Message sent, waiting for acknowledgement from server")
         waitflag = asyncio.Event()
@@ -151,14 +148,18 @@ class LWLink2:
             _LOGGER.info("async_sendmessage: Send failed, resending message (attempt %s)", _retry + 1)
             return await self._async_sendmessage(message, _retry + 1, redact)
 
-    # Use asyncio.coroutine for compatibility with Python 3.5
-    @asyncio.coroutine
-    def _consumer_handler(self):
+    async def _consumer_handler(self):
         while True:
             _LOGGER.debug("consumer_handler: Starting consumer handler")
             try:
-                mess = yield from self._websocket.receive()
+                mess = await self._websocket.receive()
                 _LOGGER.debug("consumer_handler: Received %s", mess)
+            except AttributeError:  # websocket is None if not set up, just wait for a while
+                _LOGGER.debug("consumer_handler: websocket not ready, sleeping for 1 sec")
+                await asyncio.sleep(1)
+            except Exception as exp:
+                _LOGGER.warning("consumer_handler: unhandled exception ('{}')".format(exp))
+            else:
                 if mess.type == aiohttp.WSMsgType.TEXT:
                     message = mess.json()
                     # Some transaction IDs don't work, this is a workaround
@@ -168,23 +169,23 @@ class LWLink2:
                     # now parse the message
                     if message["transactionId"] in self._transactions:
                         _LOGGER.debug("consumer_handler: Response matched for transaction %s", message["transactionId"])
+                        self._response = message
                         self._transactions[message["transactionId"]].set()
                         self._transactions.pop(message["transactionId"])
-                        self._response = message
                     elif message["direction"] == "notification" and message["class"] == "group" \
                             and message["operation"] == "event":
-                        yield from self.async_get_hierarchy()
+                        await self.async_get_hierarchy()
                     elif message["direction"] == "notification" and message["operation"] == "event":
                         if "featureId" in message["items"][0]["payload"]:
                             feature_id = message["items"][0]["payload"]["featureId"]
                             feature = self.get_feature_by_featureid(feature_id)
                             value = message["items"][0]["payload"]["value"]
-                            prev_value = self.get_featureset_by_featureid(feature_id).features[feature][1]
-                            self.get_featureset_by_featureid(feature_id).features[feature][1] = value
+                            prev_value = feature.state
+                            feature._state = value
                             cblist = [c.__name__ for c in self._callback]
                             _LOGGER.debug("consumer_handler: Event received (%s %s %s), calling callbacks %s", feature_id, feature, value, cblist)
                             for func in self._callback:
-                                func(feature=feature, feature_id=feature_id, prev_value = prev_value, new_value = value)
+                                func(feature=feature.name, feature_id=feature.id, prev_value = prev_value, new_value = value)
                         else:
                             _LOGGER.warning("consumer_handler: Unhandled event message: %s", message)
                     else:
@@ -200,11 +201,6 @@ class LWLink2:
                     self._authtoken = None
                     asyncio.ensure_future(self.async_connect())
                     _LOGGER.info("consumer_handler: Websocket reconnect requested by message handler")
-            except AttributeError:  # websocket is None if not set up, just wait for a while
-                _LOGGER.debug("consumer_handler: websocket not ready, sleeping for 1 sec")
-                yield from asyncio.sleep(1)
-            except Exception as exp:
-                _LOGGER.warning("consumer_handler: unhandled exception ('{}')".format(exp))
 
     async def async_register_callback(self, callback):
         _LOGGER.debug("async_register_callback: Register callback '%s'", callback.__name__)
@@ -246,22 +242,30 @@ class LWLink2:
                     if z not in self.featuresets:
                         _LOGGER.debug("async_read_groups: Creating device {}".format(x))
                         new_featureset = LWRFFeatureSet()
+                        new_featureset.link = self
                         new_featureset.featureset_id = z
                         new_featureset.name = x["name"]
                         self.featuresets[z] = new_featureset
 
                     _LOGGER.debug("async_read_groups: Adding device features {}".format(x))
                     y = self.featuresets[z]
-                    y.features[x["attributes"]["type"]] = [x["featureId"], 0] #Something has changed meaning the server doesn't return values on first call
+                    feature = LWRFFeature()
+                    feature.featureset = y
+                    feature.name = x["attributes"]["type"]
+                    feature.id = x["featureId"]
+                    y.features[feature.name] = feature 
 
             for x in list(response["items"][0]["payload"]["devices"].values()):
                 self.get_featureset_by_featureid(x['featureIds'][0]).product_code = x['productCode']
 
     async def async_update_featureset_states(self):
-        for dummy, x in self.featuresets.items():
-            for y in x.features:
-                value = await self.async_read_feature(x.features[y][0])
-                x.features[y][1] = value["items"][0]["payload"]["value"]
+        for x in self.featuresets.values():
+            for y in x.features.values():
+                value = await self.async_read_feature(y.id)
+                if value["items"][0]["success"] == True:
+                    y._state = value["items"][0]["payload"]["value"]
+                else:
+                    _LOGGER.warning("update_featureset_states: failed to read feature ({}), returned {}".format(x.features[y][0], value))
 
     async def async_write_feature(self, feature_id, value):
         readmess = _LWRFWebsocketMessage("feature", "write")
@@ -269,67 +273,52 @@ class LWLink2:
         readmess.additem(readitem)
         await self._async_sendmessage(readmess)
 
+    async def async_write_feature_by_name(self, featureset_id, featurename, value):
+        await self.featuresets[featureset_id].features[featurename].set_state(value)
+
     async def async_read_feature(self, feature_id):
         readmess = _LWRFWebsocketMessage("feature", "read")
         readitem = _LWRFWebsocketMessageItem({"featureId": feature_id})
         readmess.additem(readitem)
         return await self._async_sendmessage(readmess)
 
-    def get_featureset_by_id(self, featureset_id):
-        return self.featuresets[featureset_id]
-
     def get_featureset_by_featureid(self, feature_id):
-        for dummy, x in self.featuresets.items():
+        for x in self.featuresets.values():
             for y in x.features.values():
-                if y[0] == feature_id:
+                if y.id == feature_id:
                     return x
         return None
 
     def get_feature_by_featureid(self, feature_id):
-        for dummy, x in self.featuresets.items():
-            for z, y in x.features.items():
-                if y[0] == feature_id:
-                    return z
+        for x in self.featuresets.values():
+            for y in x.features.values():
+                if y.id == feature_id:
+                    return y
+        return None
 
     async def async_turn_on_by_featureset_id(self, featureset_id):
-        y = self.get_featureset_by_id(featureset_id)
-        feature_id = y.features["switch"][0]
-        await self.async_write_feature(feature_id, 1)
+        await self.async_write_feature_by_name(featureset_id, "switch", 1)
 
     async def async_turn_off_by_featureset_id(self, featureset_id):
-        y = self.get_featureset_by_id(featureset_id)
-        feature_id = y.features["switch"][0]
-        await self.async_write_feature(feature_id, 0)
+        await self.async_write_feature_by_name(featureset_id, "switch", 0)
 
     async def async_set_brightness_by_featureset_id(self, featureset_id, level):
-        y = self.get_featureset_by_id(featureset_id)
-        feature_id = y.features["dimLevel"][0]
-        await self.async_write_feature(feature_id, level)
+        await self.async_write_feature_by_name(featureset_id, "dimLevel", level)
 
     async def async_set_temperature_by_featureset_id(self, featureset_id, level):
-        y = self.get_featureset_by_id(featureset_id)
-        feature_id = y.features["targetTemperature"][0]
-        await self.async_write_feature(feature_id, int(level * 10))
+        await self.async_write_feature_by_name(featureset_id, "targetTemperature", int(level * 10))
 
     async def async_set_valvelevel_by_featureset_id(self, featureset_id, level):
-        y = self.get_featureset_by_id(featureset_id)
-        feature_id = y.features["valveLevel"][0]
-        await self.async_write_feature(feature_id, int(level * 20))
+        await self.async_write_feature_by_name(featureset_id, "valveLevel", int(level * 20))
 
     async def async_cover_open_by_featureset_id(self, featureset_id):
-        y = self.get_featureset_by_id(featureset_id)
-        feature_id = y.features["threeWayRelay"][0]
-        await self.async_write_feature(feature_id, 1)
+        await self.async_write_feature_by_name(featureset_id, "threeWayRelay", 1)
 
     async def async_cover_close_by_featureset_id(self, featureset_id):
-        y = self.get_featureset_by_id(featureset_id)
-        feature_id = y.features["threeWayRelay"][0]
-        await self.async_write_feature(feature_id, 2)
+        await self.async_write_feature_by_name(featureset_id, "threeWayRelay", 2)
 
     async def async_cover_stop_by_featureset_id(self, featureset_id):
-        y = self.get_featureset_by_id(featureset_id)
-        feature_id = y.features["threeWayRelay"][0]
-        await self.async_write_feature(feature_id, 0)
+        await self.async_write_feature_by_name(featureset_id, "threeWayRelay", 0)
 
     async def async_set_led_rgb_by_featureset_id(self, featureset_id, color):
         red = (color & int("0xFF0000", 16)) >> 16
@@ -342,9 +331,7 @@ class LWLink2:
         if blue != 0:
             blue = min(max(blue , RGB_FLOOR), 255)
         newcolor = (red << 16) + (green << 8) + blue
-        y = self.get_featureset_by_id(featureset_id)
-        feature_id = y.features["rgbColor"][0]
-        await self.async_write_feature(feature_id, newcolor)
+        await self.async_write_feature_by_name(featureset_id, "rgbColor", newcolor)
 
     def get_switches(self):
         return [(x.featureset_id, x.name) for x in self.featuresets.values() if x.is_switch()]
@@ -434,7 +421,12 @@ class LWLink2:
                 _LOGGER.debug("get_access_token_username: Received response: [contents hidden for security]")
                 #_LOGGER.debug("get_access_token_username: Received response: {}".format(await req.json()))
                 self._authtoken = (await req.json())["tokens"]["access_token"]
-
+            elif req.status == 404:
+                _LOGGER.warning("get_access_token_username: Authentication failed - if network is ok, possible wrong username/password")
+                self._authtoken = None
+            else:
+                _LOGGER.warning("get_access_token_username: Authentication failed, : status {}".format(req.status))
+                self._authtoken = None
 
     # TODO check for token expiry
     async def _get_access_token_api(self):
@@ -469,6 +461,9 @@ class LWLink2:
     def write_feature(self, feature_id, value):
         return asyncio.get_event_loop().run_until_complete(self.async_write_feature(feature_id, value))
 
+    def write_feature_by_name(self, featureset_id, featurename, value):
+        return asyncio.get_event_loop().run_until_complete(self.async_write_feature_by_name(self, featureset_id, featurename, value))
+
     def read_feature(self, feature_id):
         return asyncio.get_event_loop().run_until_complete(self.async_read_feature(feature_id))
 
@@ -634,8 +629,8 @@ class LWLink2Public(LWLink2):
 
             for featuresetid in self.featuresets:
                 for featurename in self.featuresets[featuresetid].features:
-                    if self.featuresets[featuresetid].features[featurename][0] in req:
-                        self.featuresets[featuresetid].features[featurename][1] = req[self.featuresets[featuresetid].features[featurename][0]]
+                    if self.featuresets[featuresetid].features[featurename].id in req:
+                        self.featuresets[featuresetid].features[featurename]._state = req[self.featuresets[featuresetid].features[featurename][0]]
 
     async def async_write_feature(self, feature_id, value):
         payload = {"value": value}
